//! This file is heavily inspired by the FileServer implementation in Rocket
use std::path::{Path, PathBuf};

use futures::stream::TryStreamExt;
use reqwest::{
    header::{HeaderMap, HeaderName, HeaderValue},
    Method as ReqMethod,
};
use rocket::{
    response::Redirect,
    http::uri::Segments,
    http::ext::IntoOwned,
    fs::Options,
    data::ToByteUnit,
    http::{Method, Status},
    route::{Handler, Outcome},
    Data, Request, Route,
};
use tokio_util::compat::FuturesAsyncReadCompatExt;

/// This proxy server implementation proxies requests according to prefixes in their URLs. It
/// strips the mounted prefix from the URL & replaces it with the configured prefix beforing
/// proxying the request. This allows mounting a subdirectory of a server on another subdirectory
/// on the proxy (e.g. /api/v1/posts routes to /v1/posts).
#[derive(Clone)]
pub struct ProxyServer
{
    /// Base URL of the server to proxy to. This should include the protocol (http or https), as
    /// this value is passed directly to reqwest.
    root: String,
    /// With what prefix the mount prefix should be replaced
    prefix: PathBuf,
    /// Rank of the generated routes
    rank: isize,
    /// Possible options; same as FileServer
    options: Options,
}

impl ProxyServer
{
    const DEFAULT_RANK: isize = 0;

    /// Creates a new ProxyServer object.
    pub fn new<P: AsRef<Path>>(root: &str, prefix: P, rank: isize, options: Options) -> Self
    {
        ProxyServer {
            root: String::from(root),
            prefix: prefix.as_ref().into(),
            rank,
            options,
        }
    }

    /// Convenience function to initialize a ProxyServer with the default rank & options.
    pub fn from<P: AsRef<Path>>(root: &str, prefix: P) -> Self
    {
        Self::new(root, prefix, Self::DEFAULT_RANK, Options::NormalizeDirs)
    }
}

impl Into<Vec<Route>> for ProxyServer
{
    /// Converts the handler into a list of Route objects. This is used internally by Rocket to
    /// mount the handler.
    fn into(self) -> Vec<Route>
    {
        let mut routes: Vec<Route> = Vec::new();

        static METHODS: [Method; 5] = [
            Method::Get,
            Method::Post,
            Method::Patch,
            Method::Delete,
            Method::Put,
        ];

        for method in METHODS {
            let mut route = Route::ranked(self.rank, method, "/<path..>", self.clone());
            route.name = Some(
                format!(
                    "ProxyServer: {} {}{}",
                    method.as_str(),
                    self.root.clone(),
                    self.prefix.to_str().unwrap()
                )
                .into(),
            );

            routes.push(route);
        }

        routes
    }
}

#[rocket::async_trait]
impl Handler for ProxyServer
{
    async fn handle<'r>(&self, req: &'r Request<'_>, data: Data<'r>) -> Outcome<'r>
    {
        use rocket::http::uri::fmt::Path;

        // ROCKET REQUEST -> REQWEST REQUEST

        // We first parse the path & prepend it with the prefix. If we can't parse it, we return a
        // BadRequest.
        let allow_dotfiles = self.options.contains(Options::DotFiles);
        let path = req.segments::<Segments<'_, Path>>(0..).ok()
            .and_then(|segments| segments.to_path_buf(allow_dotfiles).ok())
            .map(|path| self.prefix.join(path));

        if path.is_none() {
            return Outcome::Failure(Status::BadRequest);
        }

        let path = path.unwrap(); // Unwrap is safe because we check for is_none() beforehand
        let path_str = path.to_str();

        if path_str.is_none() {
            return Outcome::Failure(Status::BadRequest);
        }

        // After confirming we can parse the path, we check if it needs to be normalized
        if self.options.contains(Options::NormalizeDirs) && !req.uri().path().ends_with('/') {
            let normal = req.uri().map_path(|p| format!("{}/", p))
                .expect("adding a trailing slash to a known good path => valid path")
                .into_owned();

            return Outcome::from_or_forward(req, data, Redirect::permanent(normal));
        }

        // Finally, we create our new URL by combining the root, path & query
        let path_str = path_str.unwrap();

        let query_part = req
            .uri()
            .query()
            .map_or(String::from(""), |s| format!("?{}", s));

        let url = format!(
            "{}{}{}",
            self.root,
            path_str,
            query_part
        );

        // Here, we convert the Rocket headers into a HeaderMap that reqwest can use.
        let mut headers = HeaderMap::new();

        for header in req.headers().iter() {
            let name = match HeaderName::from_lowercase(header.name.into_string().as_bytes()) {
                Ok(val) => val,
                Err(_) => return Outcome::Failure(Status::BadRequest),
            };
            let value = match HeaderValue::from_str(&header.value.into_owned()) {
                Ok(val) => val,
                Err(_) => return Outcome::Failure(Status::BadRequest),
            };

            headers.insert(name, value);
        }

        // Each method needs to be matched against the other's version
        let method = match req.method() {
            Method::Delete => ReqMethod::DELETE,
            Method::Get => ReqMethod::GET,
            Method::Patch => ReqMethod::PATCH,
            Method::Post => ReqMethod::POST,
            Method::Put => ReqMethod::PUT,
            _ => todo!(), // This should never be reached as the proxy only mounts on these five routes
        };

        // And finally, the data
        // TODO don't hard-code max request size here
        static BODY_SIZE: i64 = 10;
        // TODO remove unwrap
        let body = data
            .open(BODY_SIZE.mebibytes())
            .into_bytes()
            .await;

        if body.is_err() {
            return Outcome::Failure(Status::InternalServerError);
        }

        // TODO remove unwrap
        // let body = reqwest::Body::wrap_stream(data);
        let new_res = reqwest::Client::new()
            .request(method, url)
            .headers(headers)
            .body(body.unwrap().to_vec())
            .send()
            .await;

        if new_res.is_err() {
            return Outcome::Failure(Status::InternalServerError);
        }

        let new_res = new_res.unwrap();
        let mut res_headers = new_res.headers().clone();

        let mut res_builder = rocket::Response::build();

        // Let us thank this Git repo for showing me how to do this
        // https://github.com/benkay86/async-applied/blob/master/reqwest-tokio-compat/src/main.rs
        let mut res_builder = res_builder
            .status(rocket::http::Status::new(new_res.status().as_u16()))
            .streamed_body(
                new_res
                    .bytes_stream()
                    .map_err(|e| futures::io::Error::new(futures::io::ErrorKind::Other, e))
                    .into_async_read()
                    .compat(),
            );

        // reqwest headers -> rocket headers
        for (key, value) in res_headers.iter_mut() {
            res_builder = res_builder.raw_header(
                String::from(key.clone().as_str()),
                String::from(value.clone().to_str().unwrap()),
            );
        }

        Outcome::Success(res_builder.finalize())
    }
}
